#|
This file is part of Ordo.

Copyright (C) 2025 Ray Miller

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, version 3.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.
|#

(define-module (ordo action filesystem)
  #:use-module (ice-9 binary-ports)
  #:use-module (ice-9 match)
  #:use-module (logging logger)
  #:use-module (rnrs bytevectors)
  #:use-module (srfi srfi-1) ; list utils
  #:use-module (srfi srfi-26) ; cut
  #:use-module (srfi srfi-71) ; extended let
  #:use-module (ordo action remote-cmd)
  #:use-module ((ordo connection base) #:select (with-remote-output-file))
  #:export (create-tmp-dir
            install-dir
            install-file
            file-info
            remove-file
            create-link))

(define* (file-info conn #:key path (atime? #t) (ctime? #t))
  (define (parse-stat-result s)
    (match-let* (((file-type user group . rest) (string-split s #\:))
                 ((uid gid size mode atime mtime ctime) (map string->number rest)))
      `((file-type . ,file-type)
        (user . ,user)
        (group . ,group)
        (uid . ,uid)
        (gid . ,gid)
        (size . ,size)
        (mode . ,mode)
        ,@(if atime? (list (cons 'atime atime)) '())
        (mtime . ,mtime)
        ,@(if ctime? (list (cons 'ctime ctime)) '()))))
  (let ((result rc (remote-cmd conn "stat" `("--format=%F:%U:%G:%u:%g:%s:#o%a:%X:%Y:%Z" ,path))))
    (cond
     ((zero? rc) (parse-stat-result (first result)))
     ((string-contains (first result) "No such file or directory") #f)
     (else (error (format #f "stat ~a: ~a" path (first result)))))))

(define-syntax changed-if-stat-changed
  (syntax-rules ()
    ((changed-if-stat-changed conn path expr ...)
     (let ((st-before (file-info conn #:path path #:atime? #f #:ctime? #f)))
       expr ...
       (let ((st-after (file-info conn #:path path #:atime? #f #:ctime? #f)))
         (not (equal? st-before st-after)))))))

(define* (remove-file conn #:key path (recurse? #f))
  (changed-if-stat-changed
   conn path
   (remote-cmd conn "rm" "-f" (when recurse? "-r") path
               #:check? #t)))

(define* (create-link conn #:key target link-name (symbolic? #f) (force? #t) (backup? #f))
  "Create a link to @code{target} with the name @code{link-name}."
  (changed-if-stat-changed
   conn link-name
   (remote-cmd conn "ln"
               (when symbolic? "--symbolic")
               (when force? "--force")
               (when backup? "--backup=numbered")
               target
               link-name
               #:check? #t)))

(define* (create-tmp-dir conn #:key tmpdir suffix template)
  (remote-cmd conn "mktemp" "--directory"
              (when tmpdir '("--tmpdir" tmpdir))
              (when suffix '("--suffix" suffix))
              (when template template)
              #:check? #t
              #:return car))

(define* (install-dir conn #:key path owner group mode)
  (when (integer? mode)
    (set! mode (number->string mode 8)))
  (changed-if-stat-changed
   conn path
   (remote-cmd conn "install" "--directory"
               (when owner `("--owner" ,owner))
               (when group `("--group" ,group))
               (when mode `("--mode" ,mode))
               path
               #:check? #t)))

(define (upload-tmp-file conn tmp-file)
  (lambda (input-port)
    (with-remote-output-file conn tmp-file
                             (lambda (output-port)
                               (let loop ((data (get-bytevector-some input-port)))
                                 (unless (eof-object? data)
                                   (put-bytevector output-port data)
                                   (loop (get-bytevector-some input-port))))
                               (close-port output-port)))))

(define (install-remote-file conn src dest owner group mode backup?)
  ;; If owner/group/mode is unspecified and the destination file already exists,
  ;; preserve the current ownership and mode.
  (unless (and owner group mode)
    (let ((st (file-info conn #:path dest)))
      (when st
        (set! owner (or owner (assoc-ref st 'owner)))
        (set! group (or group (assoc-ref st 'group)))
        (set! mode (or mode (assoc-ref st 'mode))))))
  (when (integer? mode)
    (set! mode (number->string mode 8)))
  (remote-cmd conn "install"
              "--compare"
              (when owner `("--owner" ,owner))
              (when group `("--group" ,group))
              (when mode `("--mode" ,mode))
              (when backup? "--backup=numbered")
              src
              dest
              #:check? #t))

(define* (install-file conn #:key path owner group (mode #o644) content local-src remote-src backup?)
  (when (not (= 1 (length (filter identity (list content local-src remote-src)))))
    (error "exactly one of #:content, #:local-src, or #:remote-src is required"))
  (changed-if-stat-changed
   conn path
   (if remote-src
       (install-remote-file conn remote-src path owner group mode backup?)
       ;; Because we might need sudo to install the remote file, we first
       ;; upload the source to a temporary file, then call @code{install-remote-file} to
       ;; install the temporary file to the target path.
       (let ((tmp-file (remote-cmd conn "mktemp" #:check? #t #:return car)))
         (dynamic-wind
           (const #t)
           (lambda ()
             (cond
              (local-src (call-with-input-file local-src (upload-tmp-file conn tmp-file)))
              ((string? content) (call-with-input-string content (upload-tmp-file conn tmp-file)))
              ((bytevector? content) (call-with-input-bytevector content (upload-tmp-file conn tmp-file)))
              (else (error "unsupported type for #:content")))
             (install-remote-file conn tmp-file path owner group mode backup?))
           (lambda ()
             (remote-cmd conn "rm" "-f" tmp-file)))))))
